import logging
import unittest

import rdflib

from ontolutils import Thing

LOG_LEVEL = logging.DEBUG

_QUERY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>

SELECT ?s
WHERE {
    ?s a  owl:Thing.
}
"""


class TestSerialization(unittest.TestCase):

    def test_jsonld(self):
        thing = Thing(id="https://example.com/123")
        json_ld: str = thing.model_dump_jsonld()
        g = rdflib.Graph()
        g.parse(data=json_ld, format='json-ld')
        res = g.query(_QUERY)
        bindings = res.bindings[0]
        self.assertEqual(bindings[rdflib.Variable("s")], rdflib.URIRef("https://example.com/123"))

    def test_ttl(self):
        thing = Thing(id="https://example.com/123")
        serialized: str = thing.serialize(format='ttl', base_uri="https://example.com/")
        expected_serialization = """@prefix owl: <http://www.w3.org/2002/07/owl#> .

<https://example.com/123> a owl:Thing .

"""
        self.assertEqual(serialized, expected_serialization)

        g = rdflib.Graph()
        g.parse(data=serialized, format='ttl')
        res = g.query(_QUERY)
        bindings = res.bindings[0]
        self.assertEqual(bindings[rdflib.Variable("s")], rdflib.URIRef("https://example.com/123"))

    def test_n3(self):
        thing = Thing(id="https://example.com/123")
        serialized: str = thing.serialize(format='n3', base_uri="https://example.com/")
        expected_serialization = """@prefix owl: <http://www.w3.org/2002/07/owl#> .

<https://example.com/123> a owl:Thing .

"""
        self.assertEqual(serialized, expected_serialization)
        g = rdflib.Graph()
        g.parse(data=serialized, format='n3')
        res = g.query(_QUERY)
        bindings = res.bindings[0]
        self.assertEqual(bindings[rdflib.Variable("s")], rdflib.URIRef("https://example.com/123"))



from .creative_work import CreativeWork, SoftwareSourceCode, SoftwareApplication
from .project import Project, ResearchProject

__all__ = ('Project',
           'ResearchProject',
           'CreativeWork',
           'SoftwareSourceCode',
           'SoftwareApplication')



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class SPDX(DefinedNamespace):
    _fail = True
    Annotation: URIRef
    AnnotationType: URIRef
    AnyLicenseInfo: URIRef
    Checksum: URIRef
    ChecksumAlgorithm: URIRef
    ConjunctiveLicenseSet: URIRef
    CreationInfo: URIRef
    CrossRef: URIRef
    DisjunctiveLicenseSet: URIRef
    ExternalDocumentRef: URIRef
    ExternalRef: URIRef
    ExtractedLicensingInfo: URIRef
    File: URIRef
    FileType: URIRef
    License: URIRef
    LicenseException: URIRef
    ListedLicense: URIRef
    ListedLicenseException: URIRef
    OrLaterOperator: URIRef
    Package: URIRef
    PackageVerificationCode: URIRef
    ReferenceCategory: URIRef
    ReferenceType: URIRef
    Relationship: URIRef
    RelationshipType: URIRef
    Review: URIRef
    SimpleLicensingInfo: URIRef
    Snippet: URIRef
    SpdxDocument: URIRef
    SpdxElement: URIRef
    SpdxItem: URIRef
    WithExceptionOperator: URIRef
    algorithm: URIRef
    annotation: URIRef
    annotationType: URIRef
    artifactOf: URIRef
    checksum: URIRef
    creationInfo: URIRef
    crossRef: URIRef
    dataLicense: URIRef
    describesPackage: URIRef
    externalDocumentRef: URIRef
    externalRef: URIRef
    fileDependency: URIRef
    fileType: URIRef
    hasExtractedLicensingInfo: URIRef
    hasFile: URIRef
    licenseConcluded: URIRef
    licenseDeclared: URIRef
    licenseException: URIRef
    licenseInfoFromFiles: URIRef
    licenseInfoInFile: URIRef
    licenseInfoInSnippet: URIRef
    member: URIRef
    packageVerificationCode: URIRef
    range: URIRef
    referenceCategory: URIRef
    referenceType: URIRef
    referencesFile: URIRef
    relatedSpdxElement: URIRef
    relationship: URIRef
    relationshipType: URIRef
    reviewed: URIRef
    snippetFromFile: URIRef
    spdxDocument: URIRef
    annotationDate: URIRef
    annotator: URIRef
    attributionText: URIRef
    checksumValue: URIRef
    contextualExample: URIRef
    copyrightText: URIRef
    created: URIRef
    creator: URIRef
    date: URIRef
    deprecatedVersion: URIRef
    description: URIRef
    documentation: URIRef
    downloadLocation: URIRef
    example: URIRef
    exceptionTextHtml: URIRef
    externalDocumentId: URIRef
    externalReferenceSite: URIRef
    extractedText: URIRef
    fileContributor: URIRef
    fileName: URIRef
    filesAnalyzed: URIRef
    isDeprecatedLicenseId: URIRef
    isFsfLibre: URIRef
    isLive: URIRef
    isOsiApproved: URIRef
    isValid: URIRef
    isWayBackLink: URIRef
    licenseComments: URIRef
    licenseExceptionId: URIRef
    licenseExceptionTemplate: URIRef
    licenseExceptionText: URIRef
    licenseId: URIRef
    licenseListVersion: URIRef
    licenseText: URIRef
    licenseTextHtml: URIRef
    match: URIRef
    name: URIRef
    noticeText: URIRef
    order: URIRef
    originator: URIRef
    packageFileName: URIRef
    packageName: URIRef
    packageVerificationCodeExcludedFile: URIRef
    packageVerificationCodeValue: URIRef
    referenceLocator: URIRef
    reviewDate: URIRef
    reviewer: URIRef
    snippetName: URIRef
    sourceInfo: URIRef
    specVersion: URIRef
    standardLicenseHeader: URIRef
    standardLicenseHeaderHtml: URIRef
    standardLicenseHeaderTemplate: URIRef
    standardLicenseTemplate: URIRef
    summary: URIRef
    supplier: URIRef
    timestamp: URIRef
    url: URIRef
    versionInfo: URIRef
    annotationType_other: URIRef
    annotationType_review: URIRef
    annotator: URIRef
    checksumAlgorithm_md2: URIRef
    checksumAlgorithm_md4: URIRef
    checksumAlgorithm_md5: URIRef
    checksumAlgorithm_md6: URIRef
    checksumAlgorithm_sha1: URIRef
    checksumAlgorithm_sha224: URIRef
    checksumAlgorithm_sha256: URIRef
    checksumAlgorithm_sha384: URIRef
    checksumAlgorithm_sha512: URIRef
    fileType_application: URIRef
    fileType_archive: URIRef
    fileType_audio: URIRef
    fileType_binary: URIRef
    fileType_documentation: URIRef
    fileType_image: URIRef
    fileType_other: URIRef
    fileType_source: URIRef
    fileType_spdx: URIRef
    fileType_text: URIRef
    fileType_video: URIRef
    noassertion: URIRef
    none: URIRef
    referenceCategory_other: URIRef
    referenceCategory_packageManager: URIRef
    referenceCategory_persistentId: URIRef
    referenceCategory_security: URIRef
    relationshipType_amendment: URIRef
    relationshipType_ancestorOf: URIRef
    relationshipType_buildDependencyOf: URIRef
    relationshipType_buildToolOf: URIRef
    relationshipType_containedBy: URIRef
    relationshipType_contains: URIRef
    relationshipType_copyOf: URIRef
    relationshipType_dataFile: URIRef
    relationshipType_dataFileOf: URIRef
    relationshipType_dependencyManifestOf: URIRef
    relationshipType_dependencyOf: URIRef
    relationshipType_dependsOn: URIRef
    relationshipType_descendantOf: URIRef
    relationshipType_describedBy: URIRef
    relationshipType_describes: URIRef
    relationshipType_devDependencyOf: URIRef
    relationshipType_devToolOf: URIRef
    relationshipType_distributionArtifact: URIRef
    relationshipType_documentation: URIRef
    relationshipType_dynamicLink: URIRef
    relationshipType_exampleOf: URIRef
    relationshipType_expandedFromArchive: URIRef
    relationshipType_fileAdded: URIRef
    relationshipType_fileDeleted: URIRef
    relationshipType_fileModified: URIRef
    relationshipType_generatedFrom: URIRef
    relationshipType_generates: URIRef
    relationshipType_hasPrerequisite: URIRef
    relationshipType_metafileOf: URIRef
    relationshipType_optionalComponentOf: URIRef
    relationshipType_optionalDependencyOf: URIRef
    relationshipType_other: URIRef
    relationshipType_packageOf: URIRef
    relationshipType_patchApplied: URIRef
    relationshipType_patchFor: URIRef
    relationshipType_prerequisiteFor: URIRef
    relationshipType_providedDependencyOf: URIRef
    relationshipType_runtimeDependencyOf: URIRef
    relationshipType_staticLink: URIRef
    relationshipType_testDependencyOf: URIRef
    relationshipType_testOf: URIRef
    relationshipType_testToolOf: URIRef
    relationshipType_testcaseOf: URIRef
    relationshipType_variantOf: URIRef
    reviewed: URIRef

    _NS = Namespace("http://spdx.org/rdf/terms#")


"""Auto-generated file. Do not edit!"""
import sys

from . import __dev
from ._iana_utils import IANA
from .codemeta import CODEMETA
from .m4i import M4I
from .obo import OBO
from .qudt_kind import QUDT_KIND
from .qudt_unit import QUDT_UNIT
from .schema import SCHEMA
from .hdf5 import HDF5
from .spdx import SPDX
from ..cache import package_user_dir

sys.path.insert(0, str(package_user_dir))



import logging
import pathlib
import unittest
from typing import Union

from pydantic import EmailStr, HttpUrl

import ontolutils
from ontolutils import __version__
from ontolutils import set_logging_level

__this_dir__ = pathlib.Path(__file__).parent

from ontolutils.classes import LangString

LOG_LEVEL = logging.DEBUG


class TestQuery(unittest.TestCase):

    def setUp(self):
        logger = logging.getLogger('ontolutils')
        self.INITIAL_LOG_LEVEL = logger.level

        set_logging_level(LOG_LEVEL)

        assert logger.level == LOG_LEVEL

        @ontolutils.namespaces(ex='http://example.org/',
                               prov='https://www.w3.org/ns/prov#')
        @ontolutils.urirefs(Organization='ex:Organization',
                            name='prov:name')
        class Organization(ontolutils.Thing):
            name: str

        @ontolutils.namespaces(prov="https://www.w3.org/ns/prov#",
                               foaf="http://xmlns.com/foaf/0.1/")
        @ontolutils.urirefs(Agent='prov:Agent',
                            mbox='foaf:mbox')
        class Agent(ontolutils.Thing):
            """Pydantic Model for https://www.w3.org/ns/prov#Agent"""
            mbox: EmailStr = None  # foaf:mbox
            orga: Union[Organization, HttpUrl] = None

        self.Agent = Agent
        self.Organization = Organization

    def test_dquery(self):
        test_data = """{"@context": {"foaf": "http://xmlns.com/foaf/0.1/", "prov": "http://www.w3.org/ns/prov#",
"schema": "http://www.w3.org/2000/01/rdf-schema#", "schema": "http://schema.org/",
"local": "http://example.org/"},
"@id": "local:testperson",
"@type": "prov:Person",
"foaf:firstName": "John",
"foaf:lastName": "Doe",
"age": 1,
"schema:affiliation": {
    "@id": "local:affiliation",
    "@type": "schema:Affiliation",
    "rdfs:label": "MyAffiliation"
    }
}"""
        res = ontolutils.dquery(
            subject="prov:Person", data=test_data,
            context={"prov": "http://www.w3.org/ns/prov#",
                     "local": "http://example.org",
                     "schema": "http://schema.org/"}
        )
        self.assertIsInstance(res, list)
        self.assertTrue(len(res) == 1)
        self.assertTrue(res[0]['@type'] == 'http://www.w3.org/ns/prov#Person')
        self.assertTrue(res[0]['@id'] == 'http://example.org/testperson')

    def test_dquery_codemeta(self):
        """Read the codemeta.json file and query for schema:SoftwareSourceCode"""
        codemeta_filename = __this_dir__ / '../codemeta.json'
        self.assertTrue(codemeta_filename.exists())
        res = ontolutils.dquery(
            subject="schema:SoftwareSourceCode",
            source=codemeta_filename,
            context={"schema": "http://schema.org/"}
        )
        self.assertIsInstance(res, list)
        self.assertTrue(len(res) == 1)
        self.assertTrue(res[0]['version'] == __version__.replace("rc", "-rc"))
        self.assertTrue('author' in res[0])
        self.assertIsInstance(res[0]['author'], list)
        self.assertEqual(res[0]['author'][0]["@id"], 'https://orcid.org/0000-0001-8729-0482')
        self.assertEqual(res[0]['author'][0]['@type'], 'http://schema.org/Person')
        self.assertEqual(res[0]['author'][0]['givenName'], 'Matthias')
        self.assertEqual(res[0]['author'][0]['affiliation']["@id"], 'https://ror.org/04t3en479')

    def test_query_get_dict(self):
        """query excepts a class or a type string"""
        agent1 = self.Agent(
            label='agent1',
        )
        agent2 = self.Agent(
            label='agent2',
        )

        agents_jsonld = ontolutils.merge_jsonld(
            [agent1.model_dump_jsonld(),
             agent2.model_dump_jsonld()]
        )

        with open(__this_dir__ / 'agent1.jsonld', 'w') as f:
            f.write(
                agents_jsonld
            )
        agents = ontolutils.dquery(
            subject='prov:Agent', source=__this_dir__ / 'agent1.jsonld',
            context={'prov': 'https://www.w3.org/ns/prov#',
                     'foaf': 'http://xmlns.com/foaf/0.1/'}
        )
        self.assertEqual(len(agents), 2)
        self.assertEqual(agents[0]['label'], 'agent1')
        self.assertEqual(agents[1]['label'], 'agent2')

        agent_load = self.Agent.from_jsonld(__this_dir__ / 'agent1.jsonld', limit=1)
        self.assertEqual(agent_load.label, LangString(value='agent1'))

        (__this_dir__ / 'agent1.jsonld').unlink(missing_ok=True)

    def test_query_multiple_classes_in_jsonld(self):
        from ontolutils.classes.utils import merge_jsonld

        agent1 = self.Agent(
            label='agent1',
        )
        agent2 = self.Agent(
            label='agent2',
        )
        merged_jsonld = merge_jsonld([agent1.model_dump_jsonld(),
                                      agent2.model_dump_jsonld()])
        with open(__this_dir__ / 'agents.jsonld', 'w') as f:
            f.write(merged_jsonld)

        agentX = self.Agent.from_jsonld(__this_dir__ / 'agents.jsonld')
        self.assertEqual(len(agentX), 2)
        self.assertEqual(agentX[0].label, LangString(value='agent1'))
        self.assertEqual(agentX[1].label, LangString(value='agent2'))

    def test_recursive_query(self):
        @ontolutils.namespaces(prov="https://www.w3.org/ns/prov#",
                               foaf="http://xmlns.com/foaf/0.1/")
        @ontolutils.urirefs(Student='prov:Student',
                            mbox='foaf:mbox',
                            age='foaf:age')
        class Student(ontolutils.Thing):
            """Pydantic Model for https://www.w3.org/ns/prov#Agent"""
            mbox: EmailStr = None  # foaf:mbox
            age: int = None

        @ontolutils.namespaces(prov="https://www.w3.org/ns/prov#",
                               foaf="http://xmlns.com/foaf/0.1/")
        @ontolutils.urirefs(SectionLeader='prov:SectionLeader',
                            hasStudent='foaf:hasStudent')
        class SectionLeader(ontolutils.Thing):
            """Pydantic Model for https://www.w3.org/ns/prov#Agent"""
            hasStudent: Student = None  # foaf:mbox

        @ontolutils.namespaces(prov="https://www.w3.org/ns/prov#",
                               foaf="http://xmlns.com/foaf/0.1/")
        @ontolutils.urirefs(Professor='prov:Professor',
                            hasSectionLeader='foaf:hasSectionLeader')
        class Professor(ontolutils.Thing):
            """Pydantic Model for https://www.w3.org/ns/prov#Agent"""
            hasSectionLeader: SectionLeader = None  # foaf:mbox

        section_leader = SectionLeader(label='section_leader',
                                       hasStudent=Student(label='student', age=30))
        prof = Professor(label='Professor',
                         hasSectionLeader=section_leader)

        with open(__this_dir__ / 'supersuperagent.json', 'w') as f:
            f.write(prof.model_dump_jsonld())

        p = Professor.from_jsonld(__this_dir__ / 'supersuperagent.json')[0]
        self.assertEqual(p.label, LangString(value='Professor'))
        self.assertEqual(p.hasSectionLeader.label, LangString(value='section_leader'))
        self.assertEqual(p.hasSectionLeader.hasStudent.label, LangString(value='student'))
        self.assertEqual(p.hasSectionLeader.hasStudent.age, 30)

    def test_query(self):
        agent = self.Agent(mbox='e@mail.com',
                           orga=str(self.Organization(id="https://example.org/myorga", name="my orga").id))
        with open(__this_dir__ / 'agent.jsonld', 'w') as f:
            json_ld_str = agent.model_dump_jsonld(context={'prov': 'https://www.w3.org/ns/prov#',
                                                           'foaf': 'http://xmlns.com/foaf/0.1/',
                                                           "ex": "https://example.org/"})
            print(json_ld_str)
            f.write(
                json_ld_str
            )
        found_agents = ontolutils.query(self.Agent, source=__this_dir__ / 'agent.jsonld')
        self.assertEqual(len(found_agents), 1)
        self.assertEqual(found_agents[0].mbox, 'e@mail.com')

        with open(__this_dir__ / 'agent.jsonld', 'w') as f:
            f.write(
                """
                {
                    "@context": {
                        "prov": "https://www.w3.org/ns/prov#",
                        "foaf": "http://xmlns.com/foaf/0.1/",
                        "local": "http:/local.org/"
                    },
                    "@graph": [
                        {
                            "@id": "local:agent1",
                            "@type": "prov:Agent",
                            "foaf:mbox": "a@mail.com"
                        },
                        {
                            "@id": "local:agent2",
                            "@type": "prov:Agent",
                            "foaf:mbox": "b@mail.com"
                        },
                        {
                            "@id": "local:agent3",
                            "@type": "prov:Agent",
                            "foaf:mbox": "c@mail.com"
                        }
                    ]
                }"""
            )
        found_agents = ontolutils.query(
            self.Agent, source=__this_dir__ / 'agent.jsonld',
            limit=2)
        self.assertEqual(len(found_agents), 2)
        self.assertEqual(found_agents[0].mbox, 'a@mail.com')
        self.assertEqual(found_agents[1].mbox, 'b@mail.com')

        # find none:

        @ontolutils.namespaces(prov="https://www.w3.org/ns/prov#",
                               foaf="http://xmlns.com/foaf/0.1/")
        @ontolutils.urirefs(Mything='prov:Mything')
        class Mything(ontolutils.Thing):
            pass

        found_Mything = ontolutils.query(Mything, source=__this_dir__ / 'agent.jsonld')
        self.assertEqual(len(found_Mything), 0)

        found_Mything = ontolutils.dquery(subject="prov:Mything",
                                          source=__this_dir__ / 'agent.jsonld',
                                          context=None
                                          )
        self.assertEqual(len(found_Mything), 0)

    def tearDown(self):
        pathlib.Path(__this_dir__ / 'agent.jsonld').unlink(missing_ok=True)
        pathlib.Path(__this_dir__ / 'agents.jsonld').unlink(missing_ok=True)
        pathlib.Path(__this_dir__ / 'superagent.json').unlink(missing_ok=True)
        pathlib.Path(__this_dir__ / 'supersuperagent.json').unlink(missing_ok=True)

        set_logging_level(self.INITIAL_LOG_LEVEL)
        assert logging.getLogger('ontolutils').level == self.INITIAL_LOG_LEVEL



"""Version library for ontolutils package. Determines the version of the package."""
try:
    from importlib.metadata import version as _version
except ImportError as e:
    raise ImportError('Most likely you have python<3.9 installed. At least 3.9 is required.') from e

__version__ = _version('ontolutils')



from pydantic import HttpUrl

from ontolutils import namespaces, urirefs, Thing as BaseThing


@namespaces(schema="https://schema.org/")
@urirefs(Thing='schema:Thing',
         description='schema:description',
         name='schema:name',
         url='schema:url')
class Thing(BaseThing):
    """schema:Thing (https://schema.org/Thing)
    The most generic type of item."""
    description: str = None
    url: HttpUrl = None
    name: str = None



import pathlib

import appdirs

package_user_dir = pathlib.Path(appdirs.user_data_dir('ontolutils'))
package_user_dir.mkdir(parents=True, exist_ok=True)


def get_cache_dir() -> pathlib.Path:
    cache_dir = package_user_dir / 'cache'
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir



import logging
import unittest

import rdflib

import ontolutils
from ontolutils import namespacelib, set_logging_level
from ontolutils.namespacelib import _iana_utils

LOG_LEVEL = logging.DEBUG


class TestNamespaces(unittest.TestCase):

    def setUp(self):
        logger = logging.getLogger('ontolutils')
        self.INITIAL_LOG_LEVEL = logger.level

        set_logging_level(LOG_LEVEL)

        assert logger.level == LOG_LEVEL

    def tearDown(self):
        set_logging_level(self.INITIAL_LOG_LEVEL)
        assert logging.getLogger('ontolutils').level == self.INITIAL_LOG_LEVEL

    def test_namespace_class_structure(self):
        from ontolutils import SCHEMA, M4I, CODEMETA, QUDT_UNIT, QUDT_KIND, OBO
        for ns in [SCHEMA, M4I, CODEMETA, QUDT_UNIT, QUDT_KIND, OBO]:
            self.assertIsInstance(ns, rdflib.namespace.DefinedNamespaceMeta)
            self.assertTrue(str(ns._NS).startswith('http'))
            self.assertIsInstance(ns.__annotations__, dict)
            self.assertTrue(len(ns.__annotations__) > 0, msg=f'No annotations found for {ns}')
            for k, v in ns.__annotations__.items():
                self.assertIsInstance(k, str)
                if k[0] != '_':
                    self.assertIsInstance(getattr(ns, k), rdflib.URIRef)

    def test_iana(self):
        self.assertEqual(namespacelib.IANA.application['zip'],
                         'https://www.iana.org/assignments/media-types/application/zip')

        self.assertEqual(namespacelib.IANA.audio['mp4'],
                         'https://www.iana.org/assignments/media-types/audio/mp4')
        self.assertEqual(namespacelib.IANA.image['png'],
                         'https://www.iana.org/assignments/media-types/image/png')

        self.assertEqual(namespacelib.IANA.text['html'],
                         'https://www.iana.org/assignments/media-types/text/html')

        self.assertEqual(namespacelib.IANA.video['mp4'],
                         'https://www.iana.org/assignments/media-types/video/mp4')

        self.assertEqual(namespacelib.IANA.model['gltf+json'],
                         'https://www.iana.org/assignments/media-types/model/gltf+json')

        self.assertEqual(namespacelib.IANA.multipart['form-data'],
                         'https://www.iana.org/assignments/media-types/multipart/form-data')

        self.assertEqual(namespacelib.IANA.application['json'],
                         'https://www.iana.org/assignments/media-types/application/json')

        self.assertEqual(namespacelib.IANA.application['ld+json'],
                         'https://www.iana.org/assignments/media-types/application/ld+json')

        self.assertEqual(namespacelib.IANA.message['http'],
                         'https://www.iana.org/assignments/media-types/message/http')

        self.assertEqual(namespacelib.IANA.font['woff'],
                         'https://www.iana.org/assignments/media-types/font/woff')

        # test download
        self.assertTrue(_iana_utils.iana_cache.is_dir())
        self.assertTrue(_iana_utils.iana_cache.exists())
        application_csv = _iana_utils.iana_cache / 'application.csv'
        application_csv.unlink(missing_ok=True)
        self.assertFalse(application_csv.exists())
        namespacelib._iana_utils.get_media_type('application')
        self.assertTrue(application_csv.exists())

    def test_m4i(self):
        self.assertIsInstance(ontolutils.M4I.Tool, rdflib.URIRef)
        self.assertIsInstance(namespacelib.M4I.Tool, rdflib.URIRef)
        self.assertEqual(str(namespacelib.M4I.Tool),
                         'http://w3id.org/nfdi4ing/metadata4ing#Tool')

        with self.assertRaises(AttributeError):
            namespacelib.M4I.Invalid

    def test_qudt_unit(self):
        self.assertIsInstance(namespacelib.QUDT_UNIT.M_PER_SEC, rdflib.URIRef)
        self.assertEqual(str(namespacelib.QUDT_UNIT.M_PER_SEC),
                         'http://qudt.org/vocab/unit/M-PER-SEC')
        with self.assertRaises(AttributeError):
            namespacelib.QUDT_UNIT.METER

    def test_qudt_kind(self):
        self.assertIsInstance(namespacelib.QUDT_KIND.Mass, rdflib.URIRef)
        self.assertEqual(str(namespacelib.QUDT_KIND.Mass),
                         'http://qudt.org/vocab/quantitykind/Mass')

    def test_rdflib(self):
        self.assertIsInstance(rdflib.PROV.Agent, rdflib.URIRef)
        self.assertEqual(str(rdflib.PROV.Agent),
                         "http://www.w3.org/ns/prov#Agent")

    def test_codemeta(self):
        self.assertIsInstance(namespacelib.CODEMETA.softwareSuggestions, rdflib.URIRef)
        self.assertEqual(str(namespacelib.CODEMETA.softwareSuggestions),
                         "https://codemeta.github.io/terms/softwareSuggestions")

    def test_schema(self):
        self.assertIsInstance(namespacelib.SCHEMA.Person, rdflib.URIRef)
        self.assertEqual(str(namespacelib.SCHEMA.Person),
                         "https://schema.org/Person")

    def test_hdf5(self):
        self.assertIsInstance(namespacelib.HDF5.File, rdflib.URIRef)
        # owl:Class
        self.assertEqual(str(namespacelib.HDF5.File),
                         "http://purl.allotrope.org/ontologies/hdf5/1.8#File")
        # Dataproperty
        self.assertEqual(str(namespacelib.HDF5.name),
                         "http://purl.allotrope.org/ontologies/hdf5/1.8#name")
        # Objectproperty
        self.assertEqual(str(namespacelib.HDF5.allocationTime),
                         "http://purl.allotrope.org/ontologies/hdf5/1.8#allocationTime")



import re

from pydantic import AnyUrl
from pydantic.functional_validators import WrapValidator
from typing_extensions import Annotated
from rdflib import URIRef
from .classes.thingmodel import ThingModel


def validate_resource_type(value, handler, info):
    def check_item(item):
        if isinstance(item, str) and re.match(r'^https?://', item):
            return str(item)
        if isinstance(item, AnyUrl):
            return str(item)
        if isinstance(item, ThingModel):
            return item
        if isinstance(item, URIRef):
            return str(item)
        field_name = getattr(info, "field_name", None)
        if field_name is not None:
            msg = f"ResourceType in field '{field_name}' must be a HTTP-URL string, a pydantic AnyUrl or a Thing object. Got: {type(item)}"
        else:
            msg = "ResourceType must be a HTTP-URL string, a pydantic AnyUrl or a Thing object."
        raise ValueError(msg)

    if isinstance(value, list):
        return [check_item(v) for v in value]
    return check_item(value)


ResourceType = Annotated[
    object,
    WrapValidator(validate_resource_type)
]


def __validate_blank_node(value: str, handler, info):
    if not isinstance(value, str):
        raise ValueError(f"Blank node must be a string, not {type(value)}")
    if value.startswith('_:'):
        return value
    raise ValueError(f"Blank node must start with _: {value}")


BlankNodeType = Annotated[str, WrapValidator(__validate_blank_node)]



from .m4i import Method, TextVariable, NumericalVariable, ProcessingStep, Tool

__all__ = ('ProcessingStep',
           'Method',
           'Tool',
           'NumericalVariable',
           'TextVariable'
           )



import json
import pathlib
from typing import Union, List, Optional

from ontolutils import namespaces, urirefs
from ontolutils.classes.utils import download_file
from pydantic import HttpUrl, field_validator, Field

from .thing import Thing


@namespaces(schema="https://schema.org/")
@urirefs(Organization='schema:Organization',
         name='schema:name')
class Organization(Thing):
    """schema:Organization (https://schema.org/Organization)"""


@namespaces(schema="https://schema.org/")
@urirefs(Person='schema:Person',
         givenName='schema:givenName',
         familyName='schema:familyName',
         email='schema:email',
         affiliation='schema:affiliation'
         )
class Person(Thing):
    """schema:Person (https://schema.org/Person)"""
    givenName: str = Field(alias="given_name")
    familyName: str = Field(alias="family_name", default=None)
    email: str = None
    affiliation: Union[Organization, List[Organization]] = Field(default=None)


@namespaces(schema="https://schema.org/")
@urirefs(CreativeWork='schema:CreativeWork',
         author='schema:author',
         abstract='schema:abstract')
class CreativeWork(Thing):
    """schema:CreativeWork (not intended to use for modeling)"""
    author: Union[Person, Organization, List[Union[Person, Organization]]] = Field(default=None)
    abstract: str  = Field(default=None)


@namespaces(schema="https://schema.org/")
@urirefs(SoftwareApplication='schema:SoftwareApplication',
         applicationCategory='schema:applicationCategory',
         downloadURL='schema:downloadURL',
         version='schema:version')
class SoftwareApplication(CreativeWork):
    """schema:SoftwareApplication (https://schema.org/SoftwareApplication)"""
    applicationCategory: Union[str, HttpUrl] = Field(default=None, alias="application_category")
    downloadURL: HttpUrl = Field(default=None, alias="download_URL")
    version: str = Field(default=None, alias="softwareVersion")

    @field_validator('applicationCategory')
    @classmethod
    def _validate_applicationCategory(cls, application_category: Union[str, HttpUrl]):
        if application_category.startswith('file:'):
            return application_category.rsplit('/', 1)[-1]
        return application_category


@namespaces(schema="https://schema.org/")
@urirefs(SoftwareSourceCode='schema:SoftwareSourceCode',
         codeRepository='schema:codeRepository',
         version='schema:version',
         applicationCategory='schema:applicationCategory')
class SoftwareSourceCode(CreativeWork):
    """Pydantic implementation of schema:SoftwareSourceCode (see https://schema.org/SoftwareSourceCode)

    .. note::

        More than the below parameters are possible but not explicitly defined here.
    """
    codeRepository: Union[HttpUrl, str] = Field(default=None, alias="code_repository")
    programmingLanguage: Optional[Union[str, List[str]]] = Field(default=None, alias="programming_language")
    applicationCategory: Union[str, HttpUrl] = Field(default=None, alias="application_category")
    version: str = Field(default=None)

    @classmethod
    def from_codemeta(cls, filename: Union[str, pathlib.Path]):
        """Create a SoftwareSourceCode instance from a codemeta.json file."""

        codemeta_context_file = download_file(
            'https://raw.githubusercontent.com/codemeta/codemeta/2.0/codemeta.jsonld',
            None)
        with open(codemeta_context_file) as f:
            codemeta_context = json.load(f)['@context']

        with open(filename, encoding='utf-8') as f:
            data = json.load(f)
        _ = data.pop('@context')
        data['@context'] = codemeta_context
        return SoftwareSourceCode.from_jsonld(data=json.dumps(data), limit=1)

    @field_validator('codeRepository')
    @classmethod
    def _validate_code_repository(cls, code_repository: Union[str, HttpUrl]):
        if not isinstance(code_repository, str):
            return code_repository
        if code_repository.startswith('git+'):
            _url = HttpUrl(code_repository.split("git+", 1)[1])
        return code_repository

    @field_validator('applicationCategory')
    @classmethod
    def _validate_applicationCategory(cls, application_category: Union[str, HttpUrl]):
        if application_category.startswith('file:'):
            return application_category.rsplit('/', 1)[-1]
        return application_category



from ontolutils import urirefs, namespaces, Thing, as_id
from ontolutils.typing import ResourceType
from pydantic import EmailStr, AnyUrl, Field, model_validator


@namespaces(foaf="http://xmlns.com/foaf/0.1/")
@urirefs(
    Agent="foaf:Agent",
    name="foaf:name",
    mbox="foaf:mbox",
    gender="foaf:gender",
    yahooChatID="foaf:yahooChatID",
    account="foaf:account",
    birthday="foaf:birthday",
    icqChatID="foaf:icqChatID",
    aimChatID="foaf:aimChatID",
    jabberID="foaf:jabberID",
    made="foaf:made",
    interest="foaf:interest",
    tipjar="foaf:tipjar",
    skypeID="foaf:skypeID",
    topic_interest="foaf:topic_interest",
    age="foaf:age",
    mbox_sha1sum="foaf:mbox_sha1sum",
    status="foaf:status",
    msnChatID="foaf:msnChatID",
    openid="foaf:openid",
    holdsAccount="foaf:holdsAccount",
    weblog="foaf:weblog",
    homepage="foaf:homepage"
)
class Agent(Thing):
    """Pydantic Modell für http://www.w3.org/ns/prov#Agent

    .. Hinweis::

        Mehr als die unten aufgeführten Parameter sind möglich, aber hier nicht explizit definiert.

    Parameter
    ---------
    gender: str = None
        Geschlecht (foaf:gender)
    yahooChatID: str = None
        Yahoo Chat-ID (foaf:yahooChatID)
    account: AnyUrl = None
        Account-URL (foaf:account)
    birthday: str = None
        Geburtstag (foaf:birthday)
    icqChatID: str = None
        ICQ Chat-ID (foaf:icqChatID)
    aimChatID: str = None
        AIM Chat-ID (foaf:aimChatID)
    jabberID: str = None
        Jabber-ID (foaf:jabberID)
    made: AnyUrl = None
        Erstellt-URL (foaf:made)
    mbox: EmailStr = None
        E-Mail-Adresse (foaf:mbox)
    interest: AnyUrl = None
        Interessens-URL (foaf:interest)
    tipjar: AnyUrl = None
        Tipjar-URL (foaf:tipjar)
    skypeID: str = None
        Skype-ID (foaf:skypeID)
    topic_interest: AnyUrl = None
        Themeninteresse-URL (foaf:topic_interest)
    age: int = None
        Alter (foaf:age)
    mbox_sha1sum: str = None
        SHA1-Summe der E-Mail (foaf:mbox_sha1sum)
    status: str = None
        Status (foaf:status)
    msnChatID: str = None
        MSN Chat-ID (foaf:msnChatID)
    openid: AnyUrl = None
        OpenID-URL (foaf:openid)
    holdsAccount: AnyUrl = None
        Account-Inhaber-URL (foaf:holdsAccount)
    weblog: AnyUrl = None
        Weblog-URL (foaf:weblog)
    homepage: ResourceType = None
        Homepage (foaf:homepage)
    """
    name: str = Field(default=None, alias="name")
    gender: str = Field(default=None, alias="gender")  # foaf:gender
    yahooChatID: str = Field(default=None, alias="yahoo_chat_id")  # foaf:yahooChatID
    account: AnyUrl = Field(default=None, alias="account")  # foaf:account
    birthday: str = Field(default=None, alias="birthday")  # foaf:birthday
    icqChatID: str = Field(default=None, alias="icq_chat_id")  # foaf:icqChatID
    aimChatID: str = Field(default=None, alias="aim_chat_id")  # foaf:aimChatID
    jabberID: str = Field(default=None, alias="jabber_id")  # foaf:jabberID
    made: AnyUrl = Field(default=None, alias="made")  # foaf:made
    mbox: EmailStr = Field(default=None, alias="email")  # foaf:mbox
    interest: AnyUrl = Field(default=None, alias="interest")  # foaf:interest
    tipjar: AnyUrl = Field(default=None, alias="tipjar")  # foaf:tipjar
    skypeID: str = Field(default=None, alias="skype_id")  # foaf:skypeID
    topic_interest: AnyUrl = Field(default=None, alias="topic_interest")  # foaf:topic_interest
    age: int = Field(default=None, alias="age")  # foaf:age
    mbox_sha1sum: str = Field(default=None, alias="mbox_sha1sum")  # foaf:mbox_sha1sum
    status: str = Field(default=None, alias="status")  # foaf:status
    msnChatID: str = Field(default=None, alias="msn_chat_id")  # foaf:msnChatID
    openid: AnyUrl = Field(default=None, alias="open_id")  # foaf:openid
    holdsAccount: AnyUrl = Field(default=None, alias="holds_account")  # foaf:holdsAccount
    weblog: AnyUrl = Field(default=None, alias="weblog")  # foaf:weblog
    homepage: ResourceType = Field(default=None, alias="homepage")  # foaf:homepage

    @model_validator(mode="before")
    def change_id(self):
        """Change the id to the downloadURL"""
        return as_id(self, "openid")


@namespaces(foaf="http://xmlns.com/foaf/0.1/")
@urirefs(Person="foaf:Person",
         firstName="foaf:firstName",
         familyName="foaf:familyName",
         lastName="foaf:lastName",
         plan="foaf:plan",
         surname="foaf:surname",
         geekcode="foaf:geekcode",
         pastProject="foaf:pastProject",
         publications="foaf:publications",
         currentProject="foaf:currentProject",
         workInfoHomepage="foaf:workInfoHomepage",
         myersBriggs="foaf:myersBriggs",
         schoolHomepage="foaf:schoolHomepage",
         img="foaf:img",
         workplaceHomepage="foaf:workplaceHomepage")
class Person(Agent):
    """Pydantic Model für http://www.w3.org/ns/prov#Person

    .. note::

        Mehr als die unten aufgeführten Parameter sind möglich, aber hier nicht explizit definiert.

    Parameter
    ---------
    firstName: str
        Vorname (foaf:firstName)
    familyName: str
        Familienname (foaf:familyName)
    lastName: str
        Nachname (foaf:lastName)
    family_name: str
        Familienname (foaf:family_name)
    plan: str
        Plan (foaf:plan)
    surname: str
        Nachname (foaf:surname)
    geekcode: str
        Geekcode (foaf:geekcode)
    pastProject: str
        Vergangenes Projekt (foaf:pastProject)
    publications: str
        Publikationen (foaf:publications)
    currentProject: str
        Aktuelles Projekt (foaf:currentProject)
    workInfoHomepage: AnyUrl
        Arbeitsinfo-Homepage (foaf:workInfoHomepage)
    myersBriggs: str
        Myers-Briggs Typ (foaf:myersBriggs)
    schoolHomepage: AnyUrl
        Schul-Homepage (foaf:schoolHomepage)
    img: AnyUrl
        Bild-URL (foaf:img)
    workplaceHomepage: AnyUrl
        Arbeitsplatz-Homepage (foaf:workplaceHomepage)
    """
    firstName: str = Field(default=None, alias="first_name")
    familyName: str = Field(default=None, alias="family_name")
    lastName: str = Field(default=None, alias="last_name")
    plan: str = Field(default=None, alias="plan")
    surname: str = Field(default=None, alias="surname")
    geekcode: str = Field(default=None, alias="geekcode")
    pastProject: str = Field(default=None, alias="past_project")
    publications: str = Field(default=None, alias="publications")
    currentProject: str = Field(default=None, alias="current_project")
    workInfoHomepage: AnyUrl = Field(default=None, alias="work_info_homepage")
    myersBriggs: str = Field(default=None, alias="myers_briggs")
    schoolHomepage: AnyUrl = Field(default=None, alias="school_homepage")
    img: AnyUrl = Field(default=None, alias="img")
    workplaceHomepage: AnyUrl = Field(default=None, alias="workplace_homepage")


@namespaces(foaf="http://xmlns.com/foaf/0.1/")
@urirefs(Organization='foaf:Organization',
         name='foaf:name')
class Organization(Agent):
    """Pydantic Model for http://www.w3.org/ns/prov#Organization

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    mbox: EmailStr = None
        Email address (foaf:mbox)
    """



import pathlib
import sys
import unittest

import requests.exceptions

import utils
from ontolutils.ex import dcat, prov, foaf
from ontolutils.ex.spdx import Checksum
from ontolutils.namespacelib.spdx import SPDX

__this_dir__ = pathlib.Path(__file__).parent


def get_python_version():
    """Get the current Python version as a tuple."""
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro


class TestDcat(utils.ClassTest):

    def test_Resource(self):
        resource1 = dcat.Resource(
            title='Resource title',
            description='Resource description',
            creator=prov.Person(first_name='John', lastName='Doe'),
            version='1.0',
            identifier='https://example.com/resource'
        )
        self.assertEqual(resource1.id, 'https://example.com/resource')
        self.assertEqual(resource1.title, 'Resource title')
        self.assertEqual(resource1.description, 'Resource description')
        self.assertIsInstance(resource1.creator, prov.Person)
        self.assertEqual(resource1.creator.firstName, 'John')
        self.assertEqual(resource1.creator.lastName, 'Doe')
        self.assertEqual(resource1.version, '1.0')
        self.assertEqual(str(resource1.identifier), 'https://example.com/resource')
        resource1.contributor = foaf.Organization(name='Example Org')
        self.assertIsInstance(resource1.contributor, foaf.Organization)
        self.assertEqual(resource1.serialize("ttl"), """@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix prov: <http://www.w3.org/ns/prov#> .

<https://example.com/resource> a dcat:Resource ;
    dcterms:contributor [ a foaf:Organization ;
            foaf:name "Example Org" ] ;
    dcterms:creator [ a prov:Person ;
            foaf:firstName "John" ;
            foaf:lastName "Doe" ] ;
    dcterms:description "Resource description" ;
    dcterms:identifier <https://example.com/resource> ;
    dcterms:title "Resource title" ;
    dcat:version "1.0" .

""")

    def test_License(self):
        license1 = "https://creativecommons.org/licenses/by/4.0/"
        resource = dcat.Resource(
            title='Resource title',
            description='Resource description',
            license=license1
        )
        self.assertEqual(str(resource.license), license1)
        self.assertEqual("""@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .

[] a dcat:Resource ;
    dcterms:description "Resource description" ;
    dcterms:license <https://creativecommons.org/licenses/by/4.0/> ;
    dcterms:title "Resource title" .

""",
                         resource.serialize("ttl"))
        dist = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            license=license1,
            checksum=Checksum(
                algorithm='SHA256',
                value='d2d2d2d2d2d2d2d2d2d2'
            )
        )
        self.assertEqual(str(dist.license), license1)
        self.assertEqual(dist.checksum.algorithm, str(SPDX.checksumAlgorithm_sha256))

        dataset = dcat.Dataset(
            title='Distribution title',
            description='Distribution description',
            license=license1,
            distribution=dist
        )
        self.assertEqual(str(dataset.license), license1)
        self.assertEqual(str(dataset.distribution.license), license1)
        print(dataset.serialize("ttl"))
        print(dataset.model_dump_jsonld())

        multi_lic = dcat.Resource(
            title='Resource title',
            description='Resource description',
            license=[license1, "https://opensource.org/licenses/MIT"]
        )
        self.assertEqual(multi_lic.license, [license1, "https://opensource.org/licenses/MIT"])

    def test_has_part(self):
        r1 = dcat.Resource(
            title='Resource 1',
            description='Resource 1 description',
            identifier='https://example.com/resource1'
        )
        r2 = dcat.Resource(
            title='Resource 2',
            description='Resource 2 description',
            identifier='https://example.com/resource2',
            has_part=r1
        )
        self.assertEqual(r2.hasPart.id, r1.id)

        r3 = dcat.Resource(
            title='Resource 3',
            description='Resource 3 description',
            identifier='https://example.com/resource3',
            has_part=[r1, r2]
        )
        self.assertEqual(len(r3.hasPart), 2)
        self.assertEqual(r3.hasPart[0].id, r1.id)
        self.assertEqual(r3.hasPart[1].id, r2.id)
        self.assertEqual("""@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .

<https://example.com/resource3> a dcat:Resource ;
    dcterms:description "Resource 3 description" ;
    dcterms:hasPart <https://example.com/resource1>,
        <https://example.com/resource2> ;
    dcterms:identifier <https://example.com/resource3> ;
    dcterms:title "Resource 3" .

<https://example.com/resource2> a dcat:Resource ;
    dcterms:description "Resource 2 description" ;
    dcterms:hasPart <https://example.com/resource1> ;
    dcterms:identifier <https://example.com/resource2> ;
    dcterms:title "Resource 2" .

<https://example.com/resource1> a dcat:Resource ;
    dcterms:description "Resource 1 description" ;
    dcterms:identifier <https://example.com/resource1> ;
    dcterms:title "Resource 1" .

""", r3.serialize("ttl"))

    def test_basic_distribution(self):
        dist = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            media_type='application/x-hdf'
        )
        self.assertEqual(dist.mediaType, 'https://www.iana.org/assignments/media-types/application/x-hdf')
        dist = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            media_type='https://www.iana.org/assignments/media-types/application/json'
        )
        self.assertEqual(dist.mediaType, 'https://www.iana.org/assignments/media-types/application/json')
        dist = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            media_type='text/csv'
        )
        self.assertEqual(dist.mediaType, 'https://www.iana.org/assignments/media-types/text/csv')


    @unittest.skipIf(condition=9 < get_python_version()[1] < 13,
                     reason="Only testing on min and max python version")
    def test_Distribution(self):
        distribution_none_downloadURL = dcat.Distribution(
            id='_:b2',
            title='Distribution title',
            description='Distribution description'
        )
        self.assertEqual(distribution_none_downloadURL.id, '_:b2')
        with self.assertRaises(ValueError):
            distribution_none_downloadURL.download()

        distribution_wrongfile = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            downloadURL='file://path/invalid.txt'
        )
        with self.assertRaises(FileNotFoundError):
            distribution_wrongfile.download()

        distribution1 = dcat.Distribution(
            title='Distribution title',
            description='Distribution description',
            creator=prov.Person(first_name='John', lastName='Doe'),
            version='1.0',
            identifier='https://example.com/distribution',
            accessURL='https://example.com/distribution',
            downloadURL='https://example.com/distribution/download'
        )
        self.assertEqual(distribution1.id, 'https://example.com/distribution')
        self.assertEqual(distribution1.title, 'Distribution title')
        self.assertEqual(distribution1.description, 'Distribution description')
        self.assertIsInstance(distribution1.creator, prov.Person)
        self.assertEqual(distribution1.creator.firstName, 'John')
        self.assertEqual(distribution1.creator.lastName, 'Doe')
        self.assertEqual(distribution1.version, '1.0')
        self.assertEqual(str(distribution1.identifier), 'https://example.com/distribution')
        self.assertEqual(str(distribution1.accessURL), 'https://example.com/distribution')
        self.assertEqual(str(distribution1.downloadURL), 'https://example.com/distribution/download')

        piv_dist = dcat.Distribution(
            downloadURL=self.test_jsonld_filename
        )
        filename = piv_dist.download(timeout=10)
        self.assertTrue(filename.exists())
        self.assertEqual(filename.name, 'piv_dataset.jsonld')
        self.assertIsInstance(filename, pathlib.Path)

        local_dist = dcat.Distribution(
            downloadURL=filename
        )
        i = 0
        i_max = 3
        while i < i_max:
            try:
                local_filename = local_dist.download(timeout=60)
                break
            except requests.exceptions.HTTPSConnection as e:
                print(e)
                i += 1
        self.assertTrue(local_filename.exists())
        self.assertEqual(local_filename.name, 'piv_dataset.jsonld')
        self.assertIsInstance(local_filename, pathlib.Path)

        filename.unlink(missing_ok=True)

    def test_Dataset(self):
        person = prov.Person(id="https://example.of/123", first_name='John', lastName='Doe')
        dataset1 = dcat.Dataset(
            title='Dataset title',
            description='Dataset description',
            creator=person,
            version='1.0',
            identifier='https://example.com/dataset',
            distribution=[
                dcat.Distribution(
                    title='Distribution title',
                    description='Distribution description',
                    identifier='https://example.com/distribution',
                    accessURL='https://example.com/distribution',
                    downloadURL='https://example.com/distribution/download'
                )
            ]
        )
        self.assertEqual(dataset1.id, 'https://example.com/dataset')
        self.assertEqual(dataset1.identifier, 'https://example.com/dataset')
        self.assertEqual(dataset1.title, 'Dataset title')
        self.assertEqual(dataset1.description, 'Dataset description')
        self.assertIsInstance(dataset1.creator, prov.Person)
        self.assertEqual(dataset1.creator.firstName, 'John')
        self.assertEqual(dataset1.creator.lastName, 'Doe')
        self.assertEqual(dataset1.version, '1.0')
        self.assertEqual(str(dataset1.identifier), 'https://example.com/dataset')
        self.assertIsInstance(dataset1.distribution[0], dcat.Distribution)
        self.assertEqual(dataset1.distribution[0].title, 'Distribution title')
        self.assertEqual(dataset1.distribution[0].description, 'Distribution description')
        self.assertEqual(str(dataset1.distribution[0].id), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].identifier), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].accessURL), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].downloadURL), 'https://example.com/distribution/download')

        ds = dcat.Dataset(
            title='Dataset title',
            description='Dataset description',
            creator=person.id,
            version='1.0',
            identifier='https://example.com/dataset',
            distribution=[
                dcat.Distribution(
                    title='Distribution title',
                    description='Distribution description',
                    identifier='https://example.com/distribution',
                    accessURL='https://example.com/distribution',
                    downloadURL='https://example.com/distribution/download'
                )
            ]
        )
        ttl = ds.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .

<https://example.com/dataset> a dcat:Dataset ;
    dcterms:creator <https://example.of/123> ;
    dcterms:description "Dataset description" ;
    dcterms:identifier <https://example.com/dataset> ;
    dcterms:title "Dataset title" ;
    dcat:distribution <https://example.com/distribution> ;
    dcat:version "1.0" .

<https://example.com/distribution> a dcat:Distribution ;
    dcterms:description "Distribution description" ;
    dcterms:identifier <https://example.com/distribution> ;
    dcterms:title "Distribution title" ;
    dcat:accessURL <https://example.com/distribution> ;
    dcat:downloadURL <https://example.com/distribution/download> .

""")

    def test_Dataset_with_foaf(self):
        person = foaf.Person(openid="http://example.com/people/johndoe",

                             first_name='John', family_name='Doe')
        self.assertEqual(person.id, 'http://example.com/people/johndoe')
        dataset1 = dcat.Dataset(
            title='Dataset title',
            description='Dataset description',
            creator=person,
            version='1.0',
            identifier='https://example.com/dataset',
            distribution=[
                dcat.Distribution(
                    title='Distribution title',
                    description='Distribution description',
                    identifier='https://example.com/distribution',
                    accessURL='https://example.com/distribution',
                    downloadURL='https://example.com/distribution/download'
                )
            ]
        )
        self.assertEqual(dataset1.id, 'https://example.com/dataset')
        self.assertEqual(dataset1.identifier, 'https://example.com/dataset')
        self.assertEqual(dataset1.title, 'Dataset title')
        self.assertEqual(dataset1.description, 'Dataset description')
        self.assertIsInstance(dataset1.creator, foaf.Person)
        self.assertEqual(dataset1.creator.firstName, 'John')
        self.assertEqual(dataset1.creator.familyName, 'Doe')
        self.assertEqual(dataset1.version, '1.0')
        self.assertEqual(str(dataset1.identifier), 'https://example.com/dataset')
        self.assertIsInstance(dataset1.distribution[0], dcat.Distribution)
        self.assertEqual(dataset1.distribution[0].title, 'Distribution title')
        self.assertEqual(dataset1.distribution[0].description, 'Distribution description')
        self.assertEqual(str(dataset1.distribution[0].id), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].identifier), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].accessURL), 'https://example.com/distribution')
        self.assertEqual(str(dataset1.distribution[0].downloadURL), 'https://example.com/distribution/download')

        dataset2 = dcat.Dataset(
            title='Dataset title',
            description='Dataset description',
            creator=person.id,
            version='1.0',
            identifier='https://example.com/dataset',
            distribution=[
                dcat.Distribution(
                    title='Distribution title',
                    description='Distribution description',
                    identifier='https://example.com/distribution',
                    accessURL='https://example.com/distribution',
                    downloadURL='https://example.com/distribution/download'
                )
            ]
        )
        self.assertEqual(str(dataset2.creator), str(person.id))

    def test_was_generated_by(self):
        from ontolutils.ex.m4i import ProcessingStep
        processing_step = ProcessingStep(id='https://example.com/processingstep/1', )
        dist = dcat.Distribution(
            id='https://example.com/distribution/1',
            was_generated_by=processing_step
        )
        print(dist.serialize(format="ttl"))

    def test_dcat_DataService(self):
        data_service = dcat.DataService(
            id="http://local.org/sqlite3",
            title="sqlite3 Database",
            endpointURL="file:///path/to/endpoint",
            servesDataset=dcat.Dataset(
                title="Table Title",
                description="An SQL Table with the name 'Table Title'",
                distribution=dcat.Distribution(
                    id="http://local.org/sqlite3/12345",
                    identifier="12345",
                    mediaType="application/vnd.sqlite3",
                )
            )
        )
        self.assertEqual(data_service.id, "http://local.org/sqlite3")
        self.assertIsInstance(data_service, dcat.DataService)



"""Ontolutils package."""
import logging

from ._cfg import set_config, get_config
from ._version import __version__
from .classes import Thing, LangString, get_urirefs, get_namespaces, as_id, build, Property, namespaces, urirefs, query, \
    dquery, URIValue
from .classes.utils import merge_jsonld
from .namespacelib import *
from .utils.qudt_units import parse_unit

DEFAULT_LOGGING_LEVEL = logging.WARNING
_formatter = logging.Formatter(
    '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d_%H:%M:%S'
)
logger = logging.getLogger('ontolutils')
_sh = logging.StreamHandler()
_sh.setFormatter(_formatter)
logger.addHandler(_sh)


def set_logging_level(level: str):
    """Set the logging level for the package and all its handlers."""
    logger.setLevel(level)
    for h in logger.handlers:
        h.setLevel(level)


set_logging_level(DEFAULT_LOGGING_LEVEL)

__all__ = ['Thing',
           'LangString',
           '__version__',
           'namespaces',
           'urirefs',
           'query',
           'set_logging_level',
           'merge_jsonld',
           'dquery',
           'get_urirefs',
           'get_namespaces',
           'parse_unit',
           'as_id',
           'build',
           'Property',
           'set_config',
           ]



import json
import logging
import pathlib
from typing import Union, List, Dict

import rdflib
import requests

logger = logging.getLogger(__package__)
logger.setLevel('DEBUG')


class UNManager:
    """Manager class for URIRef and Namespace."""

    def __init__(self):
        self.data = {}

    def get(self, cls, other=None) -> Dict:
        """Get the Namespace for the class."""
        ret = self.data.get(cls, other)
        if ret is None:
            return other
        return ret

    def __repr__(self):
        names = ', '.join([f'{c.__name__}' for c in self.data])
        return f'{self.__class__.__name__}({names})'

    def __getitem__(self, cls):
        if cls not in self.data:
            self.data[cls] = {}
        # there might be subclass to this cls. get those data as well
        for k, v in self.data.items():
            if k != cls:
                if issubclass(cls, k):
                    self.data[cls].update(v)
        return self.data[cls]


def split_URIRef(uri: rdflib.URIRef) -> List[Union[str, None]]:
    """Split a URIRef into namespace and key."""
    _uri = str(uri)
    if _uri.startswith('http'):
        if '#' in _uri:
            _split = _uri.rsplit('#', 1)
            return [f'{_split[0]}#', _split[1]]
        _split = _uri.rsplit('/', 1)
        return [f'{_split[0]}/', _split[1]]
    if ':' in _uri:
        return _uri.rsplit(':', 1)
    return [None, uri]


def merge_jsonld(jsonld_strings: List[str]) -> str:
    """Merge multiple json-ld strings into one json-ld string."""
    jsonld_dicts = [json.loads(jlds) for jlds in jsonld_strings]

    contexts = []
    for jlds in jsonld_dicts:
        if jlds['@context'] not in contexts:
            contexts.append(jlds['@context'])

    merged_contexts = {}
    for d in contexts:
        merged_contexts.update(d)

    out = {'@context': merged_contexts,
           '@graph': []}

    for jlds in jsonld_dicts:
        if '@graph' in jlds:
            out['@graph'].append(jlds['@graph'])
        else:
            data = dict(jlds.items())
            data.pop('@context')
            out['@graph'].append(data)

    return json.dumps(out, indent=2)


def download_file(url,
                  dest_filename=None,
                  known_hash=None,
                  exist_ok: bool = False,
                  **kwargs) -> pathlib.Path:
    """Download a file from a URL and check its hash
    
    Parameter
    ---------
    url: str
        The URL of the file to download
    dest_filename: str or pathlib.Path =None
        The destination filename. If None, the filename is taken from the URL
    known_hash: str
        The expected hash of the file
    exist_ok: bool
        Whether to overwrite an existing file
    **kwargs
        Additional keyword arguments passed to requests.get()
    
    Returns
    -------
    pathlib.Path
        The path to the downloaded file

    Raises
    ------
    HTTPError if the request is not successful
    ValueError if the hash of the downloaded file does not match the expected hash
    """
    if "overwrite_existing" in kwargs:
        exist_ok = kwargs.pop("overwrite_existing")
    from ..cache import get_cache_dir

    logger.debug(f'Performing request to {url}')
    response = requests.get(url, stream=True, **kwargs)
    if not response.ok:
        response.raise_for_status()

    content = response.content

    # Calculate the hash of the downloaded content
    if known_hash:
        import hashlib
        calculated_hash = hashlib.sha256(content).hexdigest()
        if not calculated_hash == known_hash:
            raise ValueError('File does not match the expected has')

    total_size = int(response.headers.get("content-length", 0))
    # block_size = 1024

    # Save the content to a file
    if dest_filename is None:
        filename = response.url.rsplit('/', 1)[1]
        dest_parent = get_cache_dir() / f'{total_size}'
        dest_filename = dest_parent / filename
        if dest_filename.exists():
            logger.debug(f'Taking existing file {dest_filename} and returning it.')
            return dest_filename
    else:
        dest_filename = pathlib.Path(dest_filename)
    dest_parent = dest_filename.parent
    if not dest_parent.exists():
        dest_parent.mkdir(parents=True)

    if dest_filename.exists():
        if exist_ok:
            logger.debug(f'Destination filename found: {dest_filename}. Deleting it, as exist_ok is True.')
            dest_filename.unlink()
        else:
            logger.debug(f'Destination filename found: {dest_filename}. Returning it')
            return dest_filename

    with open(dest_filename, "wb") as f:
        f.write(content)

    return dest_filename


def as_id(obj, field_name):
    if isinstance(obj, dict):
        return as_id_before(obj, field_name)
    raise ValueError(f"You must use mode='before' for as_id")


def as_id_before(obj: Dict, field_name: str):
    current_id = obj.get("id", None)
    if current_id is not None:
        return obj
    field_value = obj.get(field_name, None)
    if field_value is not None:
        if not str(field_value).startswith(("_:", "http")):
            logger.info(f"Field {field_name} is not a URIRef or BNode: {obj[field_name]}. "
                        f"You can only set an id for URIRefs or BNodes.")
        else:
            obj["id"] = field_value
    return obj



import unittest


from ontolutils.ex.m4i import TextVariable, NumericalVariable


class TestM4i(unittest.TestCase):

    def testTextVariable(self):
        text_variable = TextVariable(
            hasStringValue='String value',
            hasVariableDescription='Variable description'
        )
        self.assertEqual(text_variable.hasStringValue, 'String value')
        self.assertEqual(text_variable.hasVariableDescription, 'Variable description')

    def testNumericalVariableWithoutStandardName(self):
        numerical_variable = NumericalVariable(
            hasUnit='Unit',
            hasNumericalValue=1.0,
            hasMaximumValue=2.0,
            hasVariableDescription='Variable description')
        self.assertEqual(numerical_variable.hasUnit, 'Unit')
        self.assertEqual(numerical_variable.hasNumericalValue, 1.0)
        self.assertEqual(numerical_variable.hasMaximumValue, 2.0)
        self.assertEqual(numerical_variable.hasVariableDescription, 'Variable description')




from typing import Union, Optional

import rdflib
from pydantic import Field, field_validator

from ontolutils import Thing, urirefs, namespaces
from ontolutils.namespacelib.spdx import SPDX
from ontolutils.typing import ResourceType


@namespaces(spdx="http://spdx.org/rdf/terms#")
@urirefs(Checksum='spdx:Checksum',
         algorithm='spdx:algorithm',
         checksumValue='spdx:checksumValue')
class Checksum(Thing):
    """Pydantic implementation of dcat:Checksum

    Parameters
    ----------
    algorithm: str
        The algorithm used to compute the checksum (e.g. "MD5", "SHA-1", "SHA-256")
    checksumValue: str
        The checksum value
    """
    algorithm: Optional[ResourceType] = Field(default="None")  # dcat:algorithm
    checksumValue: str = Field(alias="value")  # dcat:value

    @field_validator("algorithm", mode="before")
    @classmethod
    def validate_algorithm(cls, algorithm_value: Union[str, ResourceType]) -> Union[str, ResourceType]:
        if isinstance(algorithm_value, rdflib.URIRef):
            return str(algorithm_value)
        if isinstance(algorithm_value, str) and str(algorithm_value).startswith("http"):
            return str(algorithm_value)
        if isinstance(algorithm_value, str):
            guess_name = f"checksumAlgorithm_{algorithm_value.lower()}"
            return str(SPDX[guess_name])
        return str(algorithm_value)
